/* acrostic-answer-list.c
 *
 * Copyright 2025 Tanmay Patil <tanmaynpatil105@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#include "acrostic-answer-list.h"
#include "acrostic-answer-list-row.h"

struct _AcrosticAnswerList
{
  GObject parent_object;
  GArray *obj_list;
};


static void     acrostic_answer_list_init            (AcrosticAnswerList       *self);
static void     acrostic_answer_list_class_init      (AcrosticAnswerListClass  *klass);
static void     acrostic_answer_list_model_init      (GListModelInterface *iface);
static void     acrostic_answer_list_dispose         (GObject             *object);
static GType    acrostic_answer_list_get_item_type   (GListModel          *list);
static guint    acrostic_answer_list_get_n_items     (GListModel          *list);
static gpointer acrostic_answer_list_get_item        (GListModel          *list,
                                                      guint                position);


G_DEFINE_TYPE_WITH_CODE (AcrosticAnswerList, acrostic_answer_list, G_TYPE_OBJECT,
                         G_IMPLEMENT_INTERFACE (G_TYPE_LIST_MODEL,
                                                acrostic_answer_list_model_init));


static void
acrostic_answer_list_init (AcrosticAnswerList *self)
{
  self->obj_list = g_array_new (FALSE, TRUE, sizeof (AcrosticAnswerListRow *));
}

static void
acrostic_answer_list_class_init (AcrosticAnswerListClass *klass)
{
  GObjectClass *object_class;

  object_class = G_OBJECT_CLASS (klass);

  object_class->dispose = acrostic_answer_list_dispose;
}

static void
acrostic_answer_list_model_init (GListModelInterface *iface)
{
  iface->get_item_type = acrostic_answer_list_get_item_type;
  iface->get_n_items = acrostic_answer_list_get_n_items;
  iface->get_item = acrostic_answer_list_get_item;
}

static void
acrostic_answer_list_dispose (GObject *object)
{
  AcrosticAnswerList *self;

  self = ACROSTIC_ANSWER_LIST (object);

  g_clear_pointer (&self->obj_list, g_array_unref);

  G_OBJECT_CLASS (acrostic_answer_list_parent_class)->dispose (object);
}

static GType
acrostic_answer_list_get_item_type (GListModel *model)
{
  return ACROSTIC_TYPE_ANSWER_LIST_ROW;
}

static guint
acrostic_answer_list_get_n_items (GListModel *model)
{
  AcrosticAnswerList *list_model;

  list_model = ACROSTIC_ANSWER_LIST (model);

  return list_model->obj_list->len;
}

static gpointer
acrostic_answer_list_get_item (GListModel *model,
                               guint       position)
{
  AcrosticAnswerList *list_model;
  AcrosticAnswerListRow *row;

  list_model = ACROSTIC_ANSWER_LIST (model);

  row = g_array_index (list_model->obj_list, AcrosticAnswerListRow *, position);

  return g_object_ref (row);
}

AcrosticAnswerList *
acrostic_answer_list_new (void)
{
  return (AcrosticAnswerList *) g_object_new (ACROSTIC_TYPE_ANSWER_LIST, NULL);
}

void
acrostic_answer_list_append (AcrosticAnswerList *self,
                             const gchar        *answer,
                             gboolean            selected)
{
  AcrosticAnswerListRow *row;

  row = acrostic_answer_list_row_new (answer, selected);

  g_array_append_val (self->obj_list, row);

  g_list_model_items_changed (G_LIST_MODEL (self), self->obj_list->len - 1, 0, 1);
}

void
acrostic_answer_list_clear (AcrosticAnswerList *self)
{
  guint n_answers;

  n_answers = self->obj_list->len;

  g_array_remove_range (self->obj_list, 0, n_answers);

  if (n_answers > 0)
    g_list_model_items_changed (G_LIST_MODEL (self), 0, n_answers, 0);
}

void
acrostic_answer_list_set_all_selected (AcrosticAnswerList *self,
                                       gboolean            selected)
{
  guint n_answers;

  n_answers = self->obj_list->len;

  for (guint i = 0; i < n_answers; i++)
    {
      AcrosticAnswerListRow *row;

      row = g_array_index (self->obj_list, AcrosticAnswerListRow *, i);

      acrostic_answer_list_row_set_selected (row, selected);
    }
}
