/* puzzle-render.c - Render puzzles using Cairo
 *
 * Copyright 2025 Toluwaleke Ogundipe <anonymoux47@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

/* The renderer implemented herein is used in the following parts of the
 * codebase (possibly non-exhaustive):
 *
 * - play_grid_snapshot(): to render puzzle grid overlays.
 * - xwd_thumbnail_puzzle(): to render puzzle thumbnails used by:
 *
 *   - `PuzzleButton`: for puzzle thumbnails in the game's puzzle picker.
 *   - `BasicTemplateRow`: for thumbnails of puzzle templates in the editor's greeter.
 *
 * - crosswords-thumbnailer: to render entire puzzle grids with overlays.
 *
 * Note that it is not used to render actual (interactive) puzzle grids in the
 * game or editor; the `PlayGrid` widget and its children, PlayCell` and
 * `PlayBorder`, are used for that instead.
 */


#include "color-palette.h"
#include "crosswords-config.h"
#include "puzzle-render.h"


static const GdkRGBA color_black = { PALETTE (BLACK) };

static const GdkRGBA enumeration_colors[] =
{
  [RENDER_COLOR_SCHEME_LIGHT] = { PALETTE (DARK_1) },
  [RENDER_COLOR_SCHEME_DARK]  = { PALETTE (DARK_2) },
};

static const GdkRGBA focused_overlay_colors[] =
{
  [RENDER_COLOR_SCHEME_LIGHT] = { PALETTE_DARKER (BLUE_4) },
  [RENDER_COLOR_SCHEME_DARK]  = { PALETTE_DARKER (BLUE_5) },
};

/* Keep the following border/cell color definitions in sync with style[-dark].css.
 *
 * The CSS named colors are defined in `color-palette.h`. It also defines macros
 * to help initialize color definitions. */

static const GdkRGBA border_colors_light[] =
{
  [LAYOUT_BORDER_STYLE_UNSET]         = { PALETTE_TRANSPARENT },
  [LAYOUT_BORDER_STYLE_VISIBLE_NULL]  = { PALETTE_WITH_ALPHA  (DARK_3, 0.15) },
  [LAYOUT_BORDER_STYLE_DARK]          = { PALETTE             (DARK_3) },
  [LAYOUT_BORDER_STYLE_NORMAL]        = { PALETTE             (LIGHT_4) },
  [LAYOUT_BORDER_STYLE_FOCUSED]       = { PALETTE             (BLUE_4) },
  [LAYOUT_BORDER_STYLE_SELECTED]      = { PALETTE             (YELLOW_4) },
  [LAYOUT_BORDER_STYLE_HIGHLIGHTED]   = { PALETTE             (ORANGE_2) },
  [LAYOUT_BORDER_STYLE_INITIAL_VAL]   = { PALETTE             (LIGHT_5) },
  [LAYOUT_BORDER_STYLE_GRID_SELECTED] = { PALETTE             (BLUE_5) },
};

/* Initialize only the styles defined in styles-dark.css.
 * Undefined styles fall back to their light variants. */
static const GdkRGBA border_colors_dark_def[] =
{
  [LAYOUT_BORDER_STYLE_DARK]          = { PALETTE             (DARK_4) },
  [LAYOUT_BORDER_STYLE_NORMAL]        = { PALETTE             (LIGHT_5) },
  [LAYOUT_BORDER_STYLE_FOCUSED]       = { PALETTE             (BLUE_5) },
  [LAYOUT_BORDER_STYLE_SELECTED]      = { PALETTE             (YELLOW_5) },
  [LAYOUT_BORDER_STYLE_HIGHLIGHTED]   = { PALETTE             (ORANGE_3) },
  [LAYOUT_BORDER_STYLE_INITIAL_VAL]   = { PALETTE             (DARK_1) },
  [LAYOUT_BORDER_STYLE_GRID_SELECTED] = { PALETTE             (BLUE_5) },
};

#define BORDER_COLOR_DARK_INIT(style, colors_suffix)\
  [style] = &border_colors_##colors_suffix[style]

static const GdkRGBA *border_colors_dark[G_N_ELEMENTS (border_colors_light)] =
{
  BORDER_COLOR_DARK_INIT (LAYOUT_BORDER_STYLE_UNSET         , light),
  BORDER_COLOR_DARK_INIT (LAYOUT_BORDER_STYLE_VISIBLE_NULL  , light),
  BORDER_COLOR_DARK_INIT (LAYOUT_BORDER_STYLE_DARK          , dark_def),
  BORDER_COLOR_DARK_INIT (LAYOUT_BORDER_STYLE_NORMAL        , dark_def),
  BORDER_COLOR_DARK_INIT (LAYOUT_BORDER_STYLE_FOCUSED       , dark_def),
  BORDER_COLOR_DARK_INIT (LAYOUT_BORDER_STYLE_SELECTED      , dark_def),
  BORDER_COLOR_DARK_INIT (LAYOUT_BORDER_STYLE_HIGHLIGHTED   , dark_def),
  BORDER_COLOR_DARK_INIT (LAYOUT_BORDER_STYLE_INITIAL_VAL   , dark_def),
  BORDER_COLOR_DARK_INIT (LAYOUT_BORDER_STYLE_GRID_SELECTED , dark_def),
};

static const GdkRGBA cell_colors_light[] =
{
  [LAYOUT_ITEM_STYLE_UNSET]                  = { PALETTE_TRANSPARENT },
  [LAYOUT_ITEM_STYLE_NULL]                   = { PALETTE_TRANSPARENT },
  [LAYOUT_ITEM_STYLE_VISIBLE_NULL]           = { PALETTE_WITH_ALPHA  (LIGHT_1, 0.15) },
  [LAYOUT_ITEM_STYLE_NORMAL]                 = { PALETTE             (LIGHT_1) },
  [LAYOUT_ITEM_STYLE_FOCUSED]                = { PALETTE             (BLUE_1) },
  [LAYOUT_ITEM_STYLE_BLOCK]                  = { PALETTE             (DARK_3) },
  [LAYOUT_ITEM_STYLE_FOCUSED_BLOCK]          = { PALETTE             (BROWN_3) },
  [LAYOUT_ITEM_STYLE_ERROR]                  = { PALETTE             (RED_3) },
  [LAYOUT_ITEM_STYLE_SELECTED]               = { PALETTE             (YELLOW_2) },
  [LAYOUT_ITEM_STYLE_SELECTED_FOCUSED]       = { PALETTE_DARKER      (YELLOW_2) },
  [LAYOUT_ITEM_STYLE_SELECTED_FOCUSED_BLOCK] = { PALETTE_DARKER      (BROWN_3) },
  [LAYOUT_ITEM_STYLE_SELECTED_NULL]          = { PALETTE_WITH_ALPHA  (YELLOW_2, 0.25) },
  [LAYOUT_ITEM_STYLE_SELECTED_FOCUSED_NULL]  = { PALETTE_WITH_ALPHA  (YELLOW_2, 0.5) },
  [LAYOUT_ITEM_STYLE_HIGHLIGHTED]            = { PALETTE             (ORANGE_1) },
  [LAYOUT_ITEM_STYLE_HIGHLIGHTED_SELECTED]   = { PALETTE             (YELLOW_4) },
  [LAYOUT_ITEM_STYLE_INITIAL_VAL]            = { PALETTE             (LIGHT_3) },
  /* Same as `LAYOUT_ITEM_STYLE_INITIAL_VAL`; See play-cell.c::css_class_to_string() */
  [LAYOUT_ITEM_STYLE_CLUE_BLOCK]             = { PALETTE             (LIGHT_3) },
};

/* Initialize only the styles defined in styles-dark.css.
 * Undefined styles fall back to their light variants. */
static const GdkRGBA cell_colors_dark_def[] =
{
  [LAYOUT_ITEM_STYLE_NORMAL]                 = { PALETTE             (LIGHT_3) },
  [LAYOUT_ITEM_STYLE_FOCUSED]                = { PALETTE             (BLUE_2) },
  [LAYOUT_ITEM_STYLE_BLOCK]                  = { PALETTE             (DARK_4) },
  [LAYOUT_ITEM_STYLE_FOCUSED_BLOCK]          = { PALETTE             (BROWN_4) },
  [LAYOUT_ITEM_STYLE_ERROR]                  = { PALETTE             (RED_4) },
  [LAYOUT_ITEM_STYLE_SELECTED]               = { PALETTE             (YELLOW_2) },
  [LAYOUT_ITEM_STYLE_SELECTED_FOCUSED]       = { PALETTE_DARKER      (YELLOW_2) },
  [LAYOUT_ITEM_STYLE_HIGHLIGHTED]            = { PALETTE             (ORANGE_2) },
  [LAYOUT_ITEM_STYLE_HIGHLIGHTED_SELECTED]   = { PALETTE             (YELLOW_4) },
  [LAYOUT_ITEM_STYLE_INITIAL_VAL]            = { PALETTE             (LIGHT_4) },
  /* Same as `LAYOUT_ITEM_STYLE_INITIAL_VAL`; See play-cell.c::css_class_to_string() */
  [LAYOUT_ITEM_STYLE_CLUE_BLOCK]             = { PALETTE             (LIGHT_4) },
};

#define CELL_COLOR_DARK_INIT(style, colors_suffix)\
  [style] = &cell_colors_##colors_suffix[style]

static const GdkRGBA *cell_colors_dark[G_N_ELEMENTS (cell_colors_light)] =
{
  CELL_COLOR_DARK_INIT (LAYOUT_ITEM_STYLE_UNSET                  , light),
  CELL_COLOR_DARK_INIT (LAYOUT_ITEM_STYLE_NULL                   , light),
  CELL_COLOR_DARK_INIT (LAYOUT_ITEM_STYLE_VISIBLE_NULL           , light),
  CELL_COLOR_DARK_INIT (LAYOUT_ITEM_STYLE_NORMAL                 , dark_def),
  CELL_COLOR_DARK_INIT (LAYOUT_ITEM_STYLE_FOCUSED                , dark_def),
  CELL_COLOR_DARK_INIT (LAYOUT_ITEM_STYLE_BLOCK                  , dark_def),
  CELL_COLOR_DARK_INIT (LAYOUT_ITEM_STYLE_FOCUSED_BLOCK          , dark_def),
  CELL_COLOR_DARK_INIT (LAYOUT_ITEM_STYLE_ERROR                  , dark_def),
  CELL_COLOR_DARK_INIT (LAYOUT_ITEM_STYLE_SELECTED               , dark_def),
  CELL_COLOR_DARK_INIT (LAYOUT_ITEM_STYLE_SELECTED_FOCUSED       , dark_def),
  CELL_COLOR_DARK_INIT (LAYOUT_ITEM_STYLE_SELECTED_FOCUSED_BLOCK , light),
  CELL_COLOR_DARK_INIT (LAYOUT_ITEM_STYLE_SELECTED_NULL          , light),
  CELL_COLOR_DARK_INIT (LAYOUT_ITEM_STYLE_SELECTED_FOCUSED_NULL  , light),
  CELL_COLOR_DARK_INIT (LAYOUT_ITEM_STYLE_HIGHLIGHTED            , dark_def),
  CELL_COLOR_DARK_INIT (LAYOUT_ITEM_STYLE_HIGHLIGHTED_SELECTED   , dark_def),
  CELL_COLOR_DARK_INIT (LAYOUT_ITEM_STYLE_INITIAL_VAL            , dark_def),
  CELL_COLOR_DARK_INIT (LAYOUT_ITEM_STYLE_CLUE_BLOCK             , dark_def),
};


/* Transforms the cairo context into 'layout space', such that:
 *
 * - cairo_user_to_device(0) == viewport.x
 * - cairo_user_to_device(0) == viewport.y
 * - cairo_user_to_device_distance(document_width) == viewport.width
 * - cairo_user_to_device_distance(document_height) == viewport.height
 *
 * assuming `viewport` is in device space, and has the same aspect ratio as the layout.
 * Though even if it were not in device space, we would still be rendering correctly;
 * the assumption is specifically for the equalities above.
 *
 * This transformation preserves the layout aspect ratio, centering the layout within
 * the viewport along the axis with excess space.
 */
static void
transform_to_layout_space (cairo_t              *cr,
                           const RenderViewport *viewport,
                           const LayoutGeometry *geometry)
{
  double width_ratio, height_ratio;
  double scale;
  double x_offset = 0, y_offset = 0;

  /* The same scale must be used along both axes to preserve the layout aspect ratio.
   *
   * When the ratios (of viewport to layout) along both axes are unequal, we go with
   * the lesser ratio in order to keep the entire layout within the viewport. Then we
   * center the layout within the viewport along the axis with the greater ratio. */

  width_ratio = viewport->width / geometry->document_width;
  height_ratio = viewport->height / geometry->document_height;

  if (width_ratio == height_ratio)
    scale = width_ratio;
  else if (width_ratio < height_ratio)
    {
      scale = width_ratio;
      /* Offset the layout by half the excess height to center it vertically. */
      y_offset = (viewport->height - geometry->document_height * scale) / 2.0;
    }
  else
    {
      scale = height_ratio;
      /* Offset the layout by half the excess width to center it horizontally. */
      x_offset = (viewport->width - geometry->document_width * scale) / 2.0;
    }

  cairo_translate (cr, viewport->x + x_offset, viewport->y + y_offset);
  cairo_scale (cr, scale, scale);
}

static void
draw_path (cairo_t          *cr,
           double            start_x,
           double            start_y,
           const LayoutPath *path)
{
  cairo_save (cr);

  cairo_translate (cr, start_x, start_y);
  cairo_scale (cr, path->scale_x, path->scale_y);
  cairo_rotate (cr, path->angle);

  cairo_move_to (cr, 0, 0);

  for (guint i = 0; i < path->n_ops; i++)
    {
      const LayoutPathOp *op = &path->ops[i];

      switch (op->kind)
        {
        case LAYOUT_PATH_OP_KIND_REL_MOVE:
          const LayoutPathRelMove *rel_move = &op->params.rel_move;
          cairo_rel_move_to (cr, rel_move->dx, rel_move->dy);
          break;
        case LAYOUT_PATH_OP_KIND_REL_LINE:
          const LayoutPathRelLine *rel_line = &op->params.rel_line;
          cairo_rel_line_to (cr, rel_line->dx, rel_line->dy);
          break;
        case LAYOUT_PATH_OP_KIND_REL_CURVE:
          const LayoutPathRelCurve *rel_curve = &op->params.rel_curve;
          cairo_rel_curve_to (cr,
                              rel_curve->dx1,
                              rel_curve->dy1,
                              rel_curve->dx2,
                              rel_curve->dy2,
                              rel_curve->dx,
                              rel_curve->dy);
          break;
        case LAYOUT_PATH_OP_KIND_CLOSE:
          cairo_close_path (cr);
          break;
        default:
          g_assert_not_reached ();
        }
    }

  cairo_restore (cr);
}

static void
draw_element_path (cairo_t                    *cr,
                   const LayoutElementExtents *extents)
{
  double x = extents->x;
  double y = extents->y;

  switch (extents->shape)
    {
    case LAYOUT_ELEMENT_SHAPE_CIRCLE:
      LayoutCircle circle = extents->sizing.circle;
      cairo_arc (cr, x, y, circle.radius, 0, 2 * G_PI);
      cairo_close_path (cr);
      break;

    case LAYOUT_ELEMENT_SHAPE_LINE:
      LayoutLine line = extents->sizing.line;
      cairo_move_to (cr, x, y);
      cairo_rel_line_to (cr, line.dx, line.dy);
      break;

    case LAYOUT_ELEMENT_SHAPE_RECT:
      LayoutRect rect = extents->sizing.rect;
      cairo_rectangle (cr, x, y, rect.width, rect.height);
      break;

    case LAYOUT_ELEMENT_SHAPE_ROUNDED_RECT:
      LayoutRoundedRect r = extents->sizing.rounded_rect;
      /* Draw the path in the clockwise direction. */
      cairo_move_to (cr, x + r.radius, y);
      cairo_rel_line_to (cr, r.width - 2 * r.radius, 0);
      cairo_arc (cr, x + r.width - r.radius, y + r.radius, r.radius, -G_PI_2, 0);
      cairo_rel_line_to (cr, 0, r.height - 2 * r.radius);
      cairo_arc (cr, x + r.width - r.radius, y + r.height - r.radius, r.radius, 0, G_PI_2);
      cairo_rel_line_to (cr, -(r.width - 2 * r.radius), 0);
      cairo_arc (cr, x + r.radius, y + r.height - r.radius, r.radius, G_PI_2, G_PI);
      cairo_rel_line_to (cr, 0, -(r.height - 2 * r.radius));
      cairo_arc (cr, x + r.radius, y + r.radius, r.radius, G_PI, 3 * G_PI_2);
      cairo_close_path (cr);
      break;

    case LAYOUT_ELEMENT_SHAPE_PATH:
      draw_path (cr, extents->x, extents->y, &extents->sizing.path);
      break;

    case LAYOUT_ELEMENT_SHAPE_TEXT:
      /* FIXME: To be implemented, or removed if deemed unuseful. */
      break;
    default:
      g_assert_not_reached ();
    }
}

static void
stroke (cairo_t       *cr,
        double         width,
        const GdkRGBA *color)
{
  g_assert (width > 0);

  if (color == NULL)
    color = &color_black;
  else if (color->alpha == 0.0)
    {
      cairo_new_path (cr);  /* Clear the current path. */
      return;
    }

  cairo_set_line_width (cr, width);
  cairo_set_source_rgba (cr, color->red, color->green, color->blue, color->alpha);
  cairo_stroke (cr);
}

static void
fill (cairo_t       *cr,
      const GdkRGBA *color)
{
  if (color == NULL)
    color = &color_black;
  else if (color->alpha == 0.0)
    {
      cairo_new_path (cr);  /* Clear the current path. */
      return;
    }

  cairo_set_source_rgba (cr, color->red, color->green, color->blue, color->alpha);
  cairo_fill (cr);
}

static const GdkRGBA *
get_border_color (gboolean               bg_color_set,
                  const GdkRGBA         *bg_color,
                  LayoutItemBorderStyle  border_style,
                  RenderColorScheme      color_scheme)
{
  if (bg_color_set)
    return bg_color;

  g_assert (border_style < G_N_ELEMENTS (border_colors_light));

  if (color_scheme == RENDER_COLOR_SCHEME_DARK)
    {
      bg_color = border_colors_dark[border_style];
      g_assert (bg_color != NULL);
    }
  else
    bg_color = &border_colors_light[border_style];

  return bg_color;
}

static const GdkRGBA *
get_cell_color (gboolean               bg_color_set,
                const GdkRGBA         *bg_color,
                LayoutItemCellStyle    cell_style,
                RenderColorScheme      color_scheme)
{
  if (bg_color_set &&
      (cell_style == LAYOUT_ITEM_STYLE_NORMAL ||
       cell_style == LAYOUT_ITEM_STYLE_BLOCK))
    return bg_color;

  g_assert (cell_style < G_N_ELEMENTS (cell_colors_light));

  if (color_scheme == RENDER_COLOR_SCHEME_DARK)
    {
      bg_color = cell_colors_dark[cell_style];
      g_assert (bg_color != NULL);
    }
  else
    bg_color = &cell_colors_light[cell_style];

  return bg_color;
}

static void
render_items (cairo_t           *cr,
              GridLayout        *layout,
              RenderColorScheme  color_scheme)
{
  cairo_operator_t saved_operator;

  /* Prevent gaps between items (rectangles) that touch off pixel boundaries. */
  saved_operator = cairo_get_operator (cr);
  cairo_set_operator (cr, CAIRO_OPERATOR_SATURATE);

  for (guint grid_row = 0; grid_row < layout->grid_rows; grid_row++)
    {
      for (guint grid_column = 0; grid_column < layout->grid_columns; grid_column++)
        {
          GridCoord grid_coord = { .row = grid_row, .column = grid_column };
          LayoutItemKind kind;
          LayoutElementExtents extents;
          LayoutItemBorderStyle border_style;
          LayoutItemCellStyle cell_style;
          const GdkRGBA *color = NULL;

          kind = grid_layout_get_kind (layout, grid_coord);
          extents = grid_layout_get_item_extents (layout, grid_coord);

          switch (kind)
            {
            case LAYOUT_ITEM_KIND_INTERSECTION: {
              LayoutIntersection item = grid_layout_get_intersection (layout, grid_coord);
              border_style = grid_layout_get_border_style(layout, grid_coord);
              color = get_border_color (item.bg_color_set,
                                        &item.bg_color,
                                        border_style,
                                        color_scheme);
              break;
            }

            case LAYOUT_ITEM_KIND_BORDER_HORIZONTAL: {
              LayoutBorderHorizontal item = grid_layout_get_border_horizontal (layout, grid_coord);
              border_style = grid_layout_get_border_style(layout, grid_coord);
              color = get_border_color (item.bg_color_set,
                                        &item.bg_color,
                                        border_style,
                                        color_scheme);
              break;
            }

            case LAYOUT_ITEM_KIND_BORDER_VERTICAL: {
              LayoutBorderVertical item = grid_layout_get_border_vertical (layout, grid_coord);
              border_style = grid_layout_get_border_style(layout, grid_coord);
              color = get_border_color (item.bg_color_set,
                                        &item.bg_color,
                                        border_style,
                                        color_scheme);
              break;
            }

            case LAYOUT_ITEM_KIND_CELL: {
              LayoutCell item = grid_layout_get_cell (layout, grid_coord);
              cell_style = grid_layout_get_cell(layout, grid_coord).css_class;
              color = get_cell_color (item.bg_color_set, &item.bg_color, cell_style, color_scheme);
              break;
            }

            case LAYOUT_ITEM_KIND_CLUE_BLOCK_CELL: {
              LayoutClueBlockCell item = grid_layout_get_clue_block_cell (layout, grid_coord);

              /* Clue Block Cell should only be of type IPUZ_CELL_BLOCK */
              g_assert (item.cell_type == IPUZ_CELL_BLOCK);

              color = get_cell_color (FALSE, NULL, LAYOUT_ITEM_STYLE_CLUE_BLOCK, color_scheme);
              break;
            }

            default:
              g_assert_not_reached ();
            }

          if (color->alpha != 0.0)
            {
              draw_element_path (cr, &extents);
              fill (cr, color);
            }
        }
    }

  cairo_set_operator (cr, saved_operator);
}

static void
render_overlays (cairo_t           *cr,
                 GridLayout        *layout,
                 RenderColorScheme  color_scheme)
{
  for (guint i = 0; i < layout->overlays->len; i++)
    {
      LayoutOverlay overlay;
      LayoutItemKind item_kind;
      LayoutElementExtents extents;
      LayoutItemBorderStyle border_style;
      LayoutItemCellStyle cell_style;
      LayoutElementExtents cell_extents;
      const GdkRGBA *color = NULL;

      overlay = g_array_index (layout->overlays, LayoutOverlay, i);
      item_kind = grid_layout_get_kind (layout, overlay.coord);
      extents = grid_layout_get_overlay_extents (layout, &overlay);

      switch (overlay.kind)
        {
        case LAYOUT_OVERLAY_KIND_BARRED:
          g_assert (item_kind == LAYOUT_ITEM_KIND_BORDER_HORIZONTAL ||
                    item_kind == LAYOUT_ITEM_KIND_BORDER_VERTICAL);

          border_style = grid_layout_get_border_style(layout, overlay.coord);
          if (border_style == LAYOUT_BORDER_STYLE_FOCUSED)
            color = &focused_overlay_colors[color_scheme];

          draw_element_path (cr, &extents);
          fill (cr, color);
          break;

        case LAYOUT_OVERLAY_KIND_ENUMERATION_APOSTROPHE:
        case LAYOUT_OVERLAY_KIND_ENUMERATION_SPACE:
        case LAYOUT_OVERLAY_KIND_ENUMERATION_PERIOD:
        case LAYOUT_OVERLAY_KIND_ENUMERATION_DASH:
          g_assert (item_kind == LAYOUT_ITEM_KIND_BORDER_HORIZONTAL ||
                    item_kind == LAYOUT_ITEM_KIND_BORDER_VERTICAL);

          color = &enumeration_colors[color_scheme];

          draw_element_path (cr, &extents);
          fill (cr, color);
          break;

        case LAYOUT_OVERLAY_KIND_DIVIDED_HORIZ:
        case LAYOUT_OVERLAY_KIND_DIVIDED_VERT:
        case LAYOUT_OVERLAY_KIND_DIVIDED_UP_RIGHT:
        case LAYOUT_OVERLAY_KIND_DIVIDED_UP_LEFT:
        case LAYOUT_OVERLAY_KIND_DIVIDED_PLUS:
        case LAYOUT_OVERLAY_KIND_DIVIDED_CROSS:
          g_assert (item_kind == LAYOUT_ITEM_KIND_CELL ||
                    item_kind == LAYOUT_ITEM_KIND_CLUE_BLOCK_CELL);

          if (item_kind == LAYOUT_ITEM_KIND_CELL)
            cell_style = grid_layout_get_cell (layout, overlay.coord).css_class;
          else
            cell_style = grid_layout_get_clue_block_cell (layout, overlay.coord).css_class;

          if (cell_style == LAYOUT_ITEM_STYLE_FOCUSED
              || cell_style == LAYOUT_ITEM_STYLE_FOCUSED_BLOCK
              || cell_style == LAYOUT_ITEM_STYLE_SELECTED_FOCUSED
              || cell_style == LAYOUT_ITEM_STYLE_SELECTED_FOCUSED_BLOCK)
            color = &focused_overlay_colors[color_scheme];

          cairo_save (cr);

          cell_extents = grid_layout_get_item_extents (layout, overlay.coord);
          draw_element_path (cr, &cell_extents);
          cairo_clip (cr);

          draw_element_path (cr, &extents);
          stroke (cr, layout->geometry.border_size / 2.0, color);

          cairo_restore (cr);
          break;

        case LAYOUT_OVERLAY_KIND_ARROW_RIGHT:
        case LAYOUT_OVERLAY_KIND_ARROW_RIGHT_DOWN:
        case LAYOUT_OVERLAY_KIND_ARROW_DOWN:
        case LAYOUT_OVERLAY_KIND_ARROW_DOWN_RIGHT:
        case LAYOUT_OVERLAY_KIND_ARROW_LEFT_DOWN:
        case LAYOUT_OVERLAY_KIND_ARROW_UP_RIGHT:
          g_assert (item_kind == LAYOUT_ITEM_KIND_BORDER_HORIZONTAL ||
                    item_kind == LAYOUT_ITEM_KIND_BORDER_VERTICAL);

          border_style = grid_layout_get_border_style(layout, overlay.coord);
          if (border_style == LAYOUT_BORDER_STYLE_FOCUSED)
            color = &focused_overlay_colors[color_scheme];

          draw_element_path (cr, &extents);
          fill (cr, color);
          break;

        default:
          g_assert_not_reached ();
        }
    }
}

/* Public Funtions */

void
render_layout (cairo_t              *cr,
               const RenderViewport *viewport,
               GridLayout           *layout,
               RenderColorScheme     color_scheme)
{
  g_return_if_fail (cr != NULL);
  g_return_if_fail (cairo_status (cr) == CAIRO_STATUS_SUCCESS);
  g_return_if_fail (viewport != NULL);
  g_return_if_fail (layout != NULL);

  cairo_save (cr);

  transform_to_layout_space (cr, viewport, &layout->geometry);

  render_items (cr, layout, color_scheme);
  render_overlays (cr, layout, color_scheme);

  cairo_restore (cr);
}

void
render_layout_overlays (cairo_t              *cr,
                        const RenderViewport *viewport,
                        GridLayout           *layout,
                        RenderColorScheme     color_scheme)
{
  g_return_if_fail (cr != NULL);
  g_return_if_fail (cairo_status (cr) == CAIRO_STATUS_SUCCESS);
  g_return_if_fail (viewport != NULL);
  g_return_if_fail (layout != NULL);

  cairo_save (cr);

  transform_to_layout_space (cr, viewport, &layout->geometry);

  render_overlays (cr, layout, color_scheme);

  cairo_restore (cr);
}

void
render_puzzle (cairo_t              *cr,
               const RenderViewport *viewport,
               IpuzPuzzle           *puzzle,
               RenderColorScheme     color_scheme)
{
  LayoutConfig config;

  g_return_if_fail (puzzle != NULL);

  config = layout_config_default (ipuz_puzzle_get_puzzle_kind (puzzle));

  render_puzzle_with_config (cr, viewport, puzzle, color_scheme, config);
}

void
render_puzzle_with_config (cairo_t              *cr,
                           const RenderViewport *viewport,
                           IpuzPuzzle           *puzzle,
                           RenderColorScheme     color_scheme,
                           LayoutConfig          config)
{
  GridState *state;
  GridLayout *layout;

  g_return_if_fail (puzzle != NULL);

  state = grid_state_new (IPUZ_CROSSWORD (puzzle), NULL, GRID_STATE_VIEW);
  layout = grid_layout_new (state, config);

  render_layout (cr, viewport, layout, color_scheme);

  grid_layout_free (layout);
  grid_state_free (state);
}
